// AS_S: general classes to manage solutions, moves and attributes of
// problems that can be described as assignment of objects to resources
// which includes sequencing of objects within the respective resource


#if !defined(__HotFrame_AS_S__)
#define __HotFrame_AS_S__

#include "basic.h"
#include "generic_s.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


// forward declarations

template <class C> class AS_S;
template <class C> class AS_S_N_Shift;
template <class C> class AS_S_A1;
template <class C> class AS_S_A2;
template <class C> class AS_S_I;


// ----- class AS_S -----------------------------------------------------------------

template <class C>
class AS_S 
: public Generic_S<C>
{
  protected:
    // external object numbering: 1,...
    // external numbering of resources: 1,...
    
    vector<vector<Range> > _a; // _a[2].[0] gives first object assigned to resource
                               // _a[0] stores non-assigned objects

    Range _numberOfObjects;

  public:
    static bool _s_positionIsRelevant;
    static bool positionIsRelevant( ) 
      { return _s_positionIsRelevant; }

    // todo: think about static _s_resourcesAreRelevant ?

    AS_S( Range nResources, Observer<C>* observer );

    virtual bool computeEvaluation( const Generic_S_N<C>& move,
                                    T& evaluation, T& delta );

    virtual void doMove( Generic_S_N<C>& move );

    virtual Generic_S<C>& operator=( const Generic_S<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const AS_S<C>& drhs = static_cast<const AS_S<C>&>( rhs );
        Generic_S<C>::operator=( rhs ); 
        _a = drhs._a;
        _numberOfObjects = drhs._numberOfObjects;
      }  
      return *this;
    }
    
    AS_S<C>& operator=( const AS_S<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S<C>&>( rhs ) ); 
      return *this;
    }

    Range objectOfResourceAtPosition( Range resource, Range position ) const
    {
      if
        (
          ! ( (resource>=0) && (resource<=_a.size()-1) )
          ||
          ! ( (position>=1) && (position<=_a[resource].size()) )
        )
        cout << "r=" << resource << " p=" << position << endl;
           
      assert( (resource>=0) && (resource<=_a.size()-1) );
      assert( (position>=1) && (position<=_a[resource].size()) );
      return _a[resource][position-1];
    }

    Range numberOfObjectsAssignedToResource( Range resource ) const
    {
      assert( (resource>=0) && (resource<=_a.size()-1) );
      return _a[resource].size();
    }

    Range numberOfResources( ) const
      { return _a.size()-1; }

    Range numberOfObjects( ) const
      { return _numberOfObjects; }

    vector<vector<Range> >& a( )
      { return _a; }

    virtual void addResources( Range numberOfNewResources );

    virtual void deleteResource( Range resource );

    virtual void addObject( Range object );

    virtual void deleteObject( Range object );

    virtual void deleteObjectAssignedToResource( Range object, Range resource );

    virtual void print( ostream& output ) const;
};

template<class C>
bool AS_S<C>::_s_positionIsRelevant = true;


// ----- class AS_S_N --------------------------------------------------------
/*
  abstract base class for neighbors of AS_S solutions
*/

template <class C>
class AS_S_N
: public Generic_S_N<C>
{
  public:
    enum NeighborhoodType { Constructive, Transitive, ConstructiveAndTransitive };
    static NeighborhoodType _s_neighborhoodType;

    static NeighborhoodType neighborhoodType( )
      { return _s_neighborhoodType; }
    static void setNeighborhoodType( NeighborhoodType neighborhoodType )
      { _s_neighborhoodType = neighborhoodType; }
    
    static void useConstructiveNeighborhood( )
      { _s_neighborhoodType = Constructive; }
    static void useTransitiveNeighborhood( )
      { _s_neighborhoodType = Transitive; }
    static void useConstructiveAndTransitiveNeighborhood( )
      { _s_neighborhoodType = ConstructiveAndTransitive; }

  protected:
    AS_S<C> *_s;

  public:
    // constructor
    AS_S_N( AS_S<C> *s, short depth = 1 )
      : Generic_S_N<C>( depth ), _s( s )
      { ; }

    AS_S_N( const AS_S_N& original )
      : Generic_S_N<C>( original ), _s( original._s )
      { ; }

    virtual void transform( vector<vector<Range> >& a ) const = 0;

    AS_S<C> * currentSolution( ) const
      { return _s; }

    virtual short actualDepth( ) const
      { return 1; }  // so far, everything is implemented only for depth==1

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const AS_S_N<C>& drhs = static_cast<const AS_S_N<C>&>( rhs );
        Generic_S_N<C>::operator=( rhs ); 
        _s = drhs._s;
      }  
      return *this;
    }

    AS_S_N<C>& operator=( const AS_S_N<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


template <class C>
typename AS_S_N<C>::NeighborhoodType
  AS_S_N<C>::_s_neighborhoodType = AS_S_N<C>::ConstructiveAndTransitive;


// ----- class AS_S_N_Shift --------------------------------------------------
/*
  this class manages shift neighbors (stores, evaluates, provides, ...)
*/

// todo: check AS_S_N_Shift with regard to an error in connection with random moves 

template <class C>
class AS_S_N_Shift
: public AS_S_N<C>
{
  protected:
    Range _resourceFrom;
    Range _positionFrom;
    Range _resourceTo;
    Range _positionTo;  // insert before object at position _positionTo
      // if not positionIsRelevant(), _positionTo only serves internal implementation needs
    
    virtual void evaluate( );

  public:
    // constructor
    AS_S_N_Shift( AS_S<C> *s,
			            NeighborhoodPosition itpos = FirstNeighbor,
                  short depth = 1 );  // so far: only implemented for depth = 1

    AS_S_N_Shift( AS_S<C> *s, short depth = 1 )
      : AS_S_N<C>( s, depth )
      { ; } 

    AS_S_N_Shift( const AS_S_N_Shift& original )
      : AS_S_N<C>( original ),
        _resourceFrom( original._resourceFrom ),
        _positionFrom( original._positionFrom ),
        _resourceTo( original._resourceTo ),
        _positionTo( original._positionTo )
      { assert(_depth==1); }

    virtual Generic_S_N<C>* clone() const
      { return new AS_S_N_Shift<C>( *this ); }

    virtual void operator++( );

    virtual bool isValid( ) const
      { return ( _resourceFrom >= 0 ); }

    virtual unsigned long size( ) const    
      {
        if ( currentSolution( )->positionIsRelevant() )
          return 2 * _s->numberOfObjects() * ( _s->numberOfObjects() - 1 );
        else
          return _s->numberOfObjects() * _s->numberOfResources();
      }

    virtual void transform( vector<vector<Range> >& a ) const;

    virtual void print( ostream& output ) const;

    Range resourceFrom( ) const
      { return _resourceFrom; }

    Range resourceTo( ) const
      { return _resourceTo; }

    Range positionFrom( ) const
      { return _positionFrom; }

    Range positionTo( ) const
      { return _positionTo; }

    Range objectToBeShifted( ) const
      { return _s->objectOfResourceAtPosition( _resourceFrom, _positionFrom ); }

    virtual Generic_S_N<C>& operator=( const Generic_S_N<C>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
        const AS_S_N_Shift<C>& drhs = static_cast<const AS_S_N_Shift<C>&>( rhs );
        AS_S_N<C>::operator=( rhs ); 
        _resourceFrom = drhs._resourceFrom;
        _positionFrom = drhs._positionFrom;
        _resourceTo = drhs._resourceTo;
        _positionTo = drhs._positionTo;
      }  
      return *this;
    }

    AS_S_N_Shift<C>& operator=( const AS_S_N_Shift<C>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const Generic_S_N<C>&>( rhs ) ); 
      return *this;
    }
};


// ----- class AS_S_A... ------------------------------------------------------------
/*
  this class manages attributes about solutions (e.g., extracts add/drop attributes from moves)
  - needed for all tabu search heuristics that use an attribute based tabu list
*/

template <class C>
class AS_S_A1
// use information about assignment of objects to resources (no position)
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Range _object, _resource;

  public:
    AS_S_A1( const AS_S_N<C>& move,
             unsigned int attributeIndex = 1 );
    AS_S_A1( );  // default constructor has to be provided (constructs dummy object)
    virtual bool operator==( const AS_S_N<C>& move ) const;

    static Range getNumberOfPositiveAttributes( const AS_S_N<C>& move );
    static Range getNumberOfNegativeAttributes( const AS_S_N<C>& move )
      { return getNumberOfPositiveAttributes( move ); }

    Range object( ) const
      { return _object; }
    Range resource( ) const
      { return _resource; }

    virtual void print( ostream& output ) const;
};


template <class C>
class AS_S_A2
// use information about edges between objects (predecessor/successor relations)
// resource numbers are treated negatively
// useful only if AS_S_N_Shift::positionIsRelevant()
{
  public:
    typedef typename C::T T;
    typedef typename C::Range Range;

  protected:
    Range _x, _y;  // arc (_x,_y)

  public:
    AS_S_A2( const AS_S_N<C>& move,
            unsigned int attributeIndex = 1 );
    AS_S_A2( );  // default constructor has to be provided (constructs dummy object)
    virtual bool operator==( const AS_S_N<C>& move ) const;

    static Range getNumberOfPositiveAttributes( const AS_S_N<C>& move );
    static Range getNumberOfNegativeAttributes( const AS_S_N<C>& move )
      { return getNumberOfPositiveAttributes( move ); }

    Range object( ) const
      { return _object; }
    Range resource( ) const
      { return _resource; }
    Range position( ) const
      { return _position; }

    virtual void print( ostream& output ) const;
};


// ----- class AS_S_I ----------------------------------------------------------------
/*
  this class manages information about solutions
  - possibly needed for all tabu search heuristics that use a trajectory based memory
  - here: done approximately using hash codes,
  -       we do not distinguish between different positions of some object in the same resource
*/

template <class C>
class AS_S_I
{
  public:
    typedef typename C::T  T;
    typedef typename C::Range Range;

  protected:
    unsigned long _hashCode;
    static HashCoefficients _s_hashCoefficients;

  public:
    AS_S_I( const AS_S<C>& s );
    AS_S_I( const AS_S_N<C>& move );
    
    virtual bool operator==( const AS_S_I<C>& rhs ) const
      { return ( rhs._hashCode == _hashCode ); }
    virtual bool operator<( const AS_S_I<C>& rhs ) const
      { return ( rhs._hashCode < _hashCode ); }
    
    Range getHashCode() const
      { return _hashCode; }
    virtual void print( ostream& output ) const;
};


// definition of static class elements -----------------------------------------

template<class C>
HashCoefficients AS_S_I<C>::_s_hashCoefficients;


// ----- implementation of class AS_S -----------------------------------------

template<class C>
AS_S<C>::AS_S( Range nResources, Observer<C>* observer )
: Generic_S<C>( observer ),             
  _numberOfObjects( 0 ),
  _a( nResources+1, vector<Range>( ) )
{
}


template<class C>
bool AS_S<C>::computeEvaluation( const Generic_S_N<C>& move,
                                 T& evaluation, T& delta )
{
  // this is an inefficient default evaluation, which only works for AS_S_N_Shift neighbors;
  // this should be redefined in an inheriting class to efficiently evaluate a neighhbor adaptively

  // this must be redefined when there are special data fields in the derived
  // solution class, which are affected by evaluate()

  // handling of AS_S_N moves

  const AS_S_N<C> *specificMove = dynamic_cast<const AS_S_N<C> *>(&move);
  if ( specificMove )
  {
    T oldEvaluation = _evaluation;

    vector<vector<Range> > originalAssignment( _a );
    specificMove->transform( _a );
  
    evaluate( );

    delta = _evaluation - oldEvaluation;
    evaluation = -delta;

    // restore current solution (*this)
    _a = originalAssignment;

    _evaluation = oldEvaluation;

    // return true;
    return false;  // to guarantee that special data fields in the derived class are updated 
  }
  else   
    assert( false );  // move not handled
}


template<class C>
void AS_S<C>::doMove( Generic_S_N<C>& move )
{
  // handling of AS_S_N moves

  const AS_S_N<C> *specificMove = dynamic_cast<const AS_S_N<C> *>(&move);
  if ( specificMove )
  {
    if ( _observer->active() )
    {
      _observer->move();
      _observer->getStream() << *specificMove; 
      _observer->flush();
      _observer->moveEvaluation( **specificMove ); 
      if ( specificMove->deltaInformation() )
        _observer->deltaEvaluation( specificMove->delta() ); 
    }

    specificMove->transform( _a );

    if ( specificMove->deltaInformation( ) )
      _evaluation += specificMove->delta(); 
    else
      evaluate();

    if ( _evaluation < _s_upperBound )
    {
      _s_upperBound = _evaluation;
      _observer->newUpperBound( _evaluation );
      _observer->timeWhenNewUpperBoundWasFound( round( (double(clock()))/CLOCKS_PER_SEC, 3 )  );
    }
  
    // checkEvaluation( );  // for debugging purposes, may/should be commented

    return;
  }
  else
    assert( false );  // move not handled
}


template<class C>
void AS_S<C>::addResources( Range numberOfNewResources )
{
  _a.insert( _a.end(), numberOfNewResources, vector<Range>( ) );
  evaluate( );
}


template<class C>
void AS_S<C>::deleteResource( Range resource )
{
  assert( resource > 0 );
  assert( resource <= numberOfResources() );

  // save objects that are currently assigned to the resource to be deleted
  // (i.e. put them into the dummy resource)
  
  _a[0].insert( _a[0].end(), _a[resource].begin(), _a[resource].end() );
  _a.erase( _a.begin() + resource );

  evaluate( );
}


template<class C>
void AS_S<C>::addObject( Range object )
{
  _a[0].push_back( object );
  ++_numberOfObjects;
  evaluate( );
}


template<class C>
void AS_S<C>::deleteObject( Range object )
{
  vector<Range>::iterator objectPosition;

  for ( int resource=0; resource<=_a.size()-1; ++resource )
  {
    objectPosition = find( _a[resource].begin(), _a[resource].end(), object );
    if ( objectPosition != _a[resource].end() )
    {
      _a[resource].erase( objectPosition );
      --_numberOfObjects;
      return;
    }
  }

  assert( false ); // should not happen
}


template<class C>
void AS_S<C>::deleteObjectAssignedToResource( Range object, Range resource )
{
  vector<Range>::iterator
    objectPosition = find( _a[resource].begin(), _a[resource].end(), object );
  if ( objectPosition != _a[resource].end() )
  {
    _a[resource].erase( objectPosition );
    --_numberOfObjects;
  }
  else
    assert( false ); // should not happen
}


template<class C>
void AS_S<C>::print( ostream& output ) const
{
  // format: nResources nObjects
  //         list the permutation (separated by spaces) 0
  //         evaluation

  output << numberOfResources() << " " << numberOfObjects() << "\n";

  for( Range i=0; i<=_a.size()-1; ++i )
  {
    output << i << " : ";
    for( Range j=1; j<=_a[i].size(); ++j )
      output << objectOfResourceAtPosition( i, j ) << " ";
    output << "0" << "\n";
  }
  output << _evaluation << "\n";
}


// ----- implementation of class AS_S_N_Shift ---------------------------------------

template <class C>
AS_S_N_Shift<C>
  ::AS_S_N_Shift( AS_S<C> *s,
                  NeighborhoodPosition itpos, // =FirstNeighbor,
                  short depth ) // =1
  : AS_S_N<C>( s, depth )
{ 
  if (    ( itpos == InvalidNeighbor )
       || ( _s->numberOfObjects() == 0 )
       || ( _s->numberOfResources() == 0 )
       || (    ( _s->numberOfObjects() == 1 )
            && ( _s->numberOfResources() == 1 )
            && ( _s->a()[0].empty() ) ) )
  {
    _resourceFrom = -1;
    return;
  }

  // we favor moves that assign some not yet assigned object
    
  if ( ( !( _s->a()[0].empty() ) ) && ( ! ( neighborhoodType() == Transitive ) ) ) 
  {
    if ( itpos == FirstNeighbor )
    {
      _resourceFrom = 0;
      _positionFrom = 1;
      _resourceTo = 1;
      if ( currentSolution()->positionIsRelevant() )
        _positionTo = 1;
      else
        _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
    }
    else // itpos == SomeRandomNeighbor
    {
      _resourceFrom = 0;
      _positionFrom = rand( _s->numberOfObjectsAssignedToResource( 0 ) ) + 1;
      _resourceTo = rand( _s->numberOfResources() ) + 1;
      if ( currentSolution()->positionIsRelevant() )
        _positionTo = rand( _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1 ) + 1;
      else
        _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
    }
  }
  else if ( ( _s->a()[0].empty() ) && ( neighborhoodType() == Constructive ) ) 
  // no valid move
  {
    _resourceFrom = -1;
    return;
  }
  else
  // here we know: transitive moves are valid, we do not have to consider the dummy resource,
  // and we do not have the situation "only one resource, which has only one object assigned"
  {
    if ( s->numberOfResources() == 1 )
    {
      if ( currentSolution()->positionIsRelevant() )
      {
        assert( s->numberOfObjects() >= 2 );

        // move some object to another position in first and only resource

        _resourceFrom = 1;
        _resourceTo = 1;

        if ( itpos == FirstNeighbor )
        {
          _positionFrom = 1;
          _positionTo = 3;
        }
        else // itpos == SomeRandomNeighbor
        {
          _positionFrom = rand( _s->numberOfObjectsAssignedToResource( _resourceFrom ) ) + 1;
          _positionTo = rand( _s->numberOfObjectsAssignedToResource( _resourceTo ) - 1 ) + 1;
        }
    
        if ( _positionTo >= _positionFrom )
          _positionTo += 2;
      }
      else
      // since there is only one resource and position is not relevant, there is no valid move
      {
        _resourceFrom = -1;
        return;
      }
    }
    else
    {
      if ( itpos == FirstNeighbor )
      {
        _resourceFrom = 1;
        // skip empty resources
        while(    ( _resourceFrom <= s->numberOfResources() )
               && ( s->numberOfObjectsAssignedToResource( _resourceFrom ) == 0 ) )
          ++_resourceFrom;

        if ( _resourceFrom > s->numberOfResources() )
        // no valid move
        {
          _resourceFrom = -1;
          return;
        }
        else
        {
          _positionFrom = 1;

          if ( _resourceFrom == 1  )
          {
            if (    ( s->numberOfObjectsAssignedToResource( _resourceFrom ) == 1 )
                 || ( ! currentSolution()->positionIsRelevant() ) )
            {
              _resourceTo = 2;

              if ( currentSolution()->positionIsRelevant() )
                _positionTo = 1;
              else
                _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
              
            }
            else  // s->numberOfObjectsAssignedToResource(_resourceFrom)>1  && positionIsRelevant()
            {
              _resourceTo = 1;
              _positionTo = 3;
            }
          }
          else  // _resourceFrom > 1

          {
            _resourceTo = 1;

            if ( currentSolution()->positionIsRelevant() )
              _positionTo = 1;
            else
              _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
          }
        }
      }
      else  // itpos == SomeRandomNeighbor
      {
        // select non-empty resource, select object in this resource,
        // select target resource (different from source resource),
        // and select position in target resource

        vector<Range> nonEmptyResources;

        for( Range i=1; i<=s->numberOfResources(); ++i )
          if ( s->numberOfObjectsAssignedToResource( i ) >= 1 )
            nonEmptyResources.push_back( i );

        _resourceFrom = nonEmptyResources[ rand( nonEmptyResources.size() ) ];

        _resourceTo = rand( _s->numberOfResources() - 1 ) + 1;
        if ( _resourceTo >= _resourceFrom )
          ++_resourceTo;

        _positionFrom = rand( _s->numberOfObjectsAssignedToResource( _resourceFrom ) ) + 1;
        _positionTo = rand( _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1 ) + 1;
      }
    }
  }

  evaluate();
  return;
}


template<class C>
void AS_S_N_Shift<C>::evaluate( )
{
  _deltaInformation =
    _s->computeEvaluation( *this, _evaluation, _delta );
}


template<class C>
void AS_S_N_Shift<C>::operator++( )
{
  if ( _depth >= 2 )
  {
    assert( false );  // not yet done
  }
  else
  {
    ++_positionTo;
    if (    ( _resourceFrom == _resourceTo )
         && ( ( _positionFrom == _positionTo ) || ( _positionFrom + 1 == _positionTo ) ) )
      _positionTo = _positionFrom + 2;
    if ( _positionTo > _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1 )
    {
      ++_resourceTo;

      if ( currentSolution()->positionIsRelevant() )
      {
        _positionTo = 1;
      
        if (    ( _resourceTo == _resourceFrom ) && ( _positionFrom == 1 ) )
          if ( _s->numberOfObjectsAssignedToResource( _resourceTo ) == 1 )
            ++_resourceTo;
          else
            _positionTo = 3;
      }
      else
      {
        if ( _resourceTo == _resourceFrom )
          ++_resourceTo;

        if ( _resourceTo <= _s->numberOfResources() )
          _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
      }

      if ( _resourceTo > _s->numberOfResources() )
      {
        _resourceTo = 1;

        if ( currentSolution()->positionIsRelevant() )
          _positionTo = 1;
        else
        {
          if ( _resourceTo == _resourceFrom )
            ++_resourceTo;

          if ( _resourceTo <= _s->numberOfResources() )
            _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
        }
        
        ++_positionFrom;

        if ( _positionFrom > _s->numberOfObjectsAssignedToResource( _resourceFrom ) ) 
        {
          if ( _resourceFrom == 0 )
            // terminate neighborhood, since we consider only moves that shift new objects
            // to some resources, if such a new object indeed exists
            _resourceFrom = -1;
          else
          {
            // skip empty resources
            do
              ++_resourceFrom;
            while ( ( _resourceFrom <= _s->numberOfResources() )
               && ( _s->numberOfObjectsAssignedToResource( _resourceFrom ) == 0 ) );

            if ( _resourceFrom > _s->numberOfResources() )
              // no valid move
              _resourceFrom = -1;
            else
            {
              _positionFrom = 1;
              _resourceTo = 1;

              if ( currentSolution()->positionIsRelevant() )
                _positionTo = 1;
              else
                _positionTo = _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1;
            }
          }
        }
      }
    }

    if ( _resourceFrom != -1 )
      evaluate();
  }
}


template<class C>
void AS_S_N_Shift<C>::transform( vector<vector<Range> >& a ) const
{
  if ( _resourceFrom != _resourceTo )
  {
    if ( _positionTo == _s->numberOfObjectsAssignedToResource( _resourceTo ) + 1 )
      a[_resourceTo].push_back( _s->a()[_resourceFrom][_positionFrom-1] );
    else
      _s->a()[_resourceTo].insert( _s->a()[_resourceTo].begin() + _positionTo - 1,
                                   _s->a()[_resourceFrom][_positionFrom-1] );

    _s->a()[_resourceFrom].erase( _s->a()[_resourceFrom].begin() + _positionFrom - 1 );
  }
  else
  {
    Range objectToBeShifted = _s->a()[_resourceFrom][_positionFrom-1];

    if ( _positionTo <= _positionFrom-1 )
    // shift one object to the left; to do this, shift the block inbetween to the right
    {
      for( Range i=_positionFrom-1; i>=_positionTo; --i )
        _s->a()[_resourceFrom][i] = _s->a()[_resourceFrom][i-1];
      _s->a()[_resourceFrom][_positionTo-1] = objectToBeShifted;
    }
    else if ( _positionTo >= _positionFrom+2 )
    // shift one object to the right; to do this, shift the block inbetween to the left
    {
      for( Range i=_positionFrom-1; i<=_positionTo-3; ++i )
        _s->a()[_resourceFrom][i] = _s->a()[_resourceFrom][i+1];
      _s->a()[_resourceFrom][_positionTo-2] = objectToBeShifted;
    }
    else
      assert( false );
  }  
}


template<class C>
void AS_S_N_Shift<C>::print( ostream& output ) const
{
  if ( currentSolution()->positionIsRelevant() )
    output << "shift object " << objectToBeShifted()
           << " from resource " << _resourceFrom
           << " (position " << _positionFrom
           << ") to resource " << _resourceTo
           << " (position " << _positionTo
           << ")\n";
  else
    output << "shift object " << objectToBeShifted()
           << " from resource " << _resourceFrom
           << " to resource " << _resourceTo
           << ")\n";
}


// ----- implementation of class AS_S_A1 ---------------------------------------

template<class C>
AS_S_A1<C>::AS_S_A1( const AS_S_N<C>& move,
                     unsigned int attributeIndex ) // =1 
{
  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( move.depth() == 1 );
      assert( ( attributeIndex == 1 ) || ( attributeIndex == -1 ) );

      _object = shift->objectToBeShifted();

      if ( attributeIndex == 1 )
        _resource = shift->resourceTo();
      else
        _resource = shift->resourceFrom();
    }
    return;
  }

  assert( false );  // move not handled
}


template<class C>
AS_S_A1<C>::AS_S_A1( )
// default constructor has to be provided (constructs dummy object/attribute)
: _object( 0 ), _resource( -1 )
{
  ;  
}


template<class C>
bool AS_S_A1<C>::operator==( const AS_S_N<C>& move ) const
{
  // checks iff an attribute of the move equals *this
  // i.e. the move would invert the attribute that is represented by *this

  // we assume a move of depth 1
  
  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      // we only check object to resource assignments here (no positions)
      
      return (    ( shift->resourceFrom() == _resource )
               && ( shift->objectToBeShifted() == _object ) );
    }
  }

  assert( false );
  return 0;
}


template<class C>
typename AS_S_A1<C>::Range AS_S_A1<C>::getNumberOfPositiveAttributes( const AS_S_N<C>& move )
{
  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
      return shift->actualDepth();
  }

  assert( false );
  return 0;
}


template<class C>
void AS_S_A1<C>::print( ostream& output ) const
{
  output << "object " << _object << " assigned to resource " << _resource << "\n";
}


// ----- implementation of class AS_S_A2 ---------------------------------------

template<class C>
AS_S_A2<C>::AS_S_A2( const AS_S_N<C>& move,
                     unsigned int attributeIndex ) // =1 
{
  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( shift->actualDepth() == 1 );
  
      if ( attributeIndex == 1 )
      // new arc 1
      {
        _x = ( shift->positionTo() == 1 ?
               - shift->resourceTo()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceTo(),
                                                                       shift->positionTo() - 1 ) );
        _y = shift->objectToBeShifted();
      }    
      else if ( attributeIndex == 2 )
      // new arc 2
      {
        _x = shift->objectToBeShifted();
        _y = ( shift->positionTo() == shift->currentSolution()
               ->numberOfObjectsAssignedToResource( shift->resourceTo() ) + 1 ) ?
               - shift->resourceTo()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceTo(),
                                                                       shift->positionTo() );
      }    
      else if ( attributeIndex == 3 )
      // new arc 3
      {
        _x = ( shift->positionFrom() == 1 ?
               - shift->resourceFrom()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(),
                                                                       shift->positionFrom() - 1 ) );
        _y = ( shift->positionFrom() == shift->currentSolution()
               ->numberOfObjectsAssignedToResource( shift->resourceFrom() ) ) ?
               - shift->resourceFrom()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(), 
                                                                       shift->positionFrom() + 1 );
      }    
      else if ( attributeIndex == -1 )
      // deleted arc 1
      {
        _x = ( shift->positionFrom() == 1 ?
               - shift->resourceFrom()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(),
                                                                       shift->positionFrom() - 1 ) );
        _y = shift->objectToBeShifted();
      }    
      else if ( attributeIndex == -2 )
      // deleted arc 2
      {
        _x = shift->objectToBeShifted();
        _y = ( shift->positionFrom() == shift->currentSolution()
               ->numberOfObjectsAssignedToResource( shift->resourceFrom() ) ) ?
               - shift->resourceFrom()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(), 
                                                                       shift->positionFrom() + 1 );
      }    
      else if ( attributeIndex == -3 )
      // deleted arc 3
      {
        _x = ( shift->positionTo() == 1 ?
               - shift->resourceTo()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceTo(),
                                                                       shift->positionTo() - 1 ) );
        _y = ( shift->positionTo() == shift->currentSolution()
               ->numberOfObjectsAssignedToResource( shift->resourceTo() ) + 1 ) ?
               - shift->resourceTo()
               : shift->currentSolution()->objectOfResourceAtPosition( shift->resourceTo(),
                                                                       shift->positionTo() );
      }    
      else
        assert( false );  // attributeIndex not handled 

      return;
    }
  }

  assert( false );  // move not handled
}


template<class C>
AS_S_A2<C>::AS_S_A2( )
// default constructor has to be provided (constructs dummy object/attribute)
: _x( 0 ), _y( 0 )
{
  ;  
}


template<class C>
bool AS_S_A2<C>::operator==( const AS_S_N<C>& move ) const
{
  // checks iff an attribute of the move equals *this
  // i.e. the move would invert the attribute that is represented by *this

  // we assume a move of depth 1
  
  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( shift->positionIsRelevant() );  // otherwise the following does not make sense
      
      bool dummy = 
        shift->positionTo() == shift->currentSolution()
                                ->numberOfObjectsAssignedToResource( shift->resourceTo() ) + 1;
      
      return (
               (    ( _x == ( shift->positionFrom() == 1 ?
                              - shift->resourceFrom()
                             : shift->currentSolution()->objectOfResourceAtPosition(
                                                             shift->resourceFrom(),
                                                             shift->positionFrom() - 1 ) ) )
                 && ( _y == shift->objectToBeShifted() )
               ) 
               ||
               (    ( _x == shift->objectToBeShifted() )
                 && ( _y == ( shift->positionFrom() == shift->currentSolution()
                                ->numberOfObjectsAssignedToResource( shift->resourceFrom() ) ?
                              - shift->resourceFrom()
                              : shift->currentSolution()->objectOfResourceAtPosition(
                                                             shift->resourceFrom(), 
                                                             shift->positionFrom() + 1 ) ) )
               )
               ||
               (    ( _x == ( shift->positionTo() == 1 ?
                              - shift->resourceTo()
                              : shift->currentSolution()->objectOfResourceAtPosition(
                                                             shift->resourceTo(),
                                                             shift->positionTo() - 1 ) ) )
                 && ( _y == ( shift->positionTo() == shift->currentSolution()
                                ->numberOfObjectsAssignedToResource( shift->resourceTo() ) + 1 ?
                              - shift->resourceTo()
                              : shift->currentSolution()->objectOfResourceAtPosition(
                                                             shift->resourceTo(),
                                                             shift->positionTo() ) ) )
               )
             );
     }
  }

  assert( false );
  return 0;
}


template<class C>
typename AS_S_A2<C>::Range AS_S_A2<C>::getNumberOfPositiveAttributes( const AS_S_N<C>& move )
{
  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      assert( move.actualDepth() == 1 );
      return 3;
    }
  }

  assert( false );
  return 0;
}


template<class C>
void AS_S_A2<C>::print( ostream& output ) const
{
  output << "arc (" << _x << "," << _y << ")" << "\n";
}


// ----- implementation of class AS_S_I ---------------------------------------

template <class C>
AS_S_I<C>::AS_S_I( const AS_S<C>& s )
{
  if ( _s_hashCoefficients.size() == 0 )
    _s_hashCoefficients.initialize( 7*s.numberOfResources() + 100,
                                    237 );

  _hashCode = 0;
  for( Range j=1; j<=s.numberOfResources(); ++j )
    for( Range k=1; k<=s.numberOfObjectsAssignedToResource( j ); ++k )
      _hashCode += s.objectOfResourceAtPosition( j, k )
                   * _s_hashCoefficients.get_coefficient( 7*j ); // i.e. position should not matter
}


template <class C>
AS_S_I<C>::AS_S_I( const AS_S_N<C>& move )
{
  if ( _s_hashCoefficients.size() == 0 )
    _s_hashCoefficients.initialize( 7*move.currentSolution()->numberOfResources() + 100,
                                    237 );

  // TODO: do the following more efficiently

  // handling of different kinds of moves

  {
    const AS_S_N_Shift<C> *shift = dynamic_cast<const AS_S_N_Shift<C> *>(&move);
    if ( shift )
    {
      _hashCode = 0;
      for( Range j=1; j<=move.currentSolution()->numberOfResources(); ++j )
        for( Range k=1; k<=move.currentSolution()->numberOfObjectsAssignedToResource( j ); ++k )
          _hashCode += move.currentSolution()->objectOfResourceAtPosition( j, k )
                       * _s_hashCoefficients.get_coefficient( 7*j );

      _hashCode -= shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(),
                                                                         shift->positionFrom() )
                   * _s_hashCoefficients.get_coefficient( 7*shift->resourceFrom() );
      _hashCode += shift->currentSolution()->objectOfResourceAtPosition( shift->resourceFrom(),
                                                                         shift->positionFrom() )
                   * _s_hashCoefficients.get_coefficient( 7*shift->resourceTo() );

      return;
    }
  }

  assert( false );
}


template<class C>
void AS_S_I<C>::print( ostream& output ) const
{
  output << _hashCode << "\n";
}


// stream output functions -----------------------------------------------------
// (all of these just call the respective (virtual) member function)

template <class C>
ostream& operator<<( ostream& output, const AS_S<C>& solution )
{
  solution.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const AS_S_N_Shift<C>& neighbor )
{
  neighbor.print( output );
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const AS_S_A1<C>& attribute )
{
  attribute.print( output );  
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const AS_S_A2<C>& attribute )
{
  attribute.print( output );  
  return output;
}

template <class C>
ostream& operator<<( ostream& output, const AS_S_I<C>& solutionInfo )
{
  solutionInfo.print( output );  
  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
