// basic.h: provides various basic functionality which is used by various hotframe components

#if !defined(__HotFrame_BASIC__)
#define __HotFrame_BASIC__


// include some files of the c++ standard library

#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <math.h>

#include <iostream>
#include <fstream>
#include <sstream>

#include <vector>
#include <list>
#include <set>
#include <map>
#include <queue>

#include <algorithm>
#include <numeric>
#include <limits>
#include <functional>

#include <string>

#include <assert.h>


using namespace std;



//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{


enum NeighborhoodPosition { FirstNeighbor, SomeRandomNeighbor, InvalidNeighbor };


// ----------------------------------------------------------------------------
// -- some useful template functions ------------------------------------------
// ----------------------------------------------------------------------------

// abs function

template <class T>
inline T abs( T x )
  { return x < 0 ? -x : x; }



// the following function templates are defined in a similar manner in stdlib,
// but there are some conflicts with microsoft's min/max in windwef.h;
// this leads to some strange errors, so they are defined here as a workaround


bool invert( bool b )
  { return b == true ? false : true; }


template <class T>
inline void checkForEqualityAndModifyInCase( T& x, T y )
{
  // [ugly workaround]
  // if x seems to equal y, but does not exactly equal x, there may be a rounding error

  if ( abs ( x - y ) < 50*numeric_limits<T>::epsilon() )
    x = y;
}


template <class T>
inline bool fuzzilyEquals( T x, T y )
{
  // [ugly workaround]
  // check whether x seems to equal y, but does not equal x,
  //   as there may be a rounding error, so return true in case

  return ( abs ( x - y ) < 5000*numeric_limits<T>::epsilon() );
}


template <class T>
ostream& operator<<( ostream& output, const vector<T>& v )
{
  for( int i=1; i<=v.size()-1; ++i )
    output << v[i] << " ";
  output << "\n";

  return output;
}



// ----------------------------------------------------------------------------
// -- simple random number generators -----------------------------------------
// ----------------------------------------------------------------------------

// TODO: do this better?, check for properties of the random number generators

inline unsigned long rand( unsigned long limit )
// provides some integer value x, 0 <= x, < limit-1
{
  if ( limit == 0 )
    return 0;
  else
    return ::rand() % limit;
}

inline double rand01( void )
// provides some double value x, 0 <= x, < 1
{
  return ((double) ::rand() ) / (RAND_MAX+1);
}



// ----------------------------------------------------------------------------
// -- simple math routines ----------------------------------------------------
// ----------------------------------------------------------------------------

inline double round( double value, int decimalPlace = 2 )
// round value to decimalPlace
{
  int i;
  
  for ( i=1; i<=decimalPlace; ++i )
    value = value * 10;

  if ( value - floor( value ) < 0.5 )
    value = floor( value );
  else
    value = ceil( value );

  for ( i=1; i<=decimalPlace; ++i )
    value = value / 10;

  return value;
}



// ----------------------------------------------------------------------------
// -- class Pair --------------------------------------------------------------
// ----------------------------------------------------------------------------

template <class T1, class T2>
class Pair
// like class pair, but with more constructors (the compiler complained ...)
{
  public:
    T1 _first;
    T2 _second;

    Pair( const T1 & first, const T2 & second )
    : _first( first ), _second( second )  { ; }

    Pair( const T1 & first )
    : _first( first ), _second( 0 )  { ; }

    Pair( )
    : _first( 0 ), _second( 0 )  { ; }

    T1 first( ) const
      { return _first; }
    T2 second( ) const
      { return _second; }

    bool operator<( const Pair<T1, T2>& rhs ) const
      { return _first < rhs._first; }
};



// ----------------------------------------------------------------------------
// -- class HashCoefficients --------------------------------------------------
// ----------------------------------------------------------------------------

class HashCoefficients
{
  protected:
    vector<unsigned long> _coefficients;

  public:
    void initialize( unsigned long n, unsigned long maxValue );

    unsigned long get_coefficient( unsigned long i ) const
      { 
        if ( i > (_coefficients.size()-1) )
          i = i % _coefficients.size();
        return _coefficients[i];
      }

    unsigned long size( ) const
      { return static_cast<unsigned long>( _coefficients.size() ); }
};

void HashCoefficients::initialize( unsigned long n, unsigned long maxValue )
{
  _coefficients.reserve( n+1 );

  for( unsigned long i=0; i<=n; ++i )
    _coefficients.push_back( ( ::rand() % maxValue ) + 1 );
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
