// Generic_S_NCL: classes representing candidate list neighborhoods for 
//                solutions that are derived from generic_s

// (this code seems to work, but it is seldomly used and thus not well tested)


#if !defined(__HotFrame_Generic_S_NCL__)
#define __HotFrame_Generic_S_NCL__

#include "basic.h"
#include "generic_s.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{

template <class N>
class betterEvaluation
{
  public:
    bool operator() ( const N& lhs, const N& rhs ) const
      { return ( *lhs > *rhs ); }
};


// -- Generic_S_NCLE -----------------------------------------------------------
/*
   general mechanism for adapting all neighborhood classes derived from Generic_S_N
   - elite strategy
   idea: one full neighborhood sweep is followed by several sweeps restricted
         to some of the best neighbors of the last full sweep;
         this might be a reasonable strategy in circumstances when the neighborhood
         evaluation is very costly and it is adequate to use the information from
         one neighbodhood sweep to perform a number of moves
*/

// forward declaration
template <class S, class N, class CNumeric> class Generic_S_NCLE;


// CLQuota (<1): the best CLQuota*neighborhoodSize neighbors are put into the candidate list
// CLPercentage(<=1): CLPercentage*candidateListSize neighborhood sweeps are done using
//                    the candidate list, followed by one full sweep
// reevaluate: do not (or do) use the old neighbor evaluation for candidate list neighbors


template <class S, class N, class CNumeric>
class Generic_S_NCLE
: public N
{
  public:
    typedef typename CNumeric::T T;
    typedef typename CNumeric::Range Range;

  protected:
    // if candidate list is active, neighbors are taken from _candidateList
    // otherwise, neighbors are constructed the normal way and put into the _candidateList
    static vector<N> _s_candidateList;
    static Range _s_candidateListPosition;
    static Range _s_remainingCandidateListTraversals;

    // parameters
    static double _s_clQuota;
    static double _s_clPercentage;
    static bool _s_reevaluate;

  public:
    // constructor
    Generic_S_NCLE( S *currentSolution,
 		                NeighborhoodPosition itpos = FirstNeighbor,
                    short depth = 1 );

    Generic_S_NCLE( const Generic_S_NCLE<S, N, CNumeric>& original )
      : N( original )
    { ; }

    virtual Generic_S_N<CNumeric>* clone() const
      { return new Generic_S_NCLE<S, N, CNumeric>( *this ); }

    virtual void operator++( );

    void reset( ) 
    {
      _s_remainingCandidateListTraversals = 0;
      _s_candidateList.clear( );
    }

    static bool candidateListActive( )
      { return ( _s_remainingCandidateListTraversals > 0 ); }

    static void setCLQuota( double CLQuota )
      { _s_clQuota = CLQuota; }
    static void setCLPercentage( double CLPercentage )
      { _s_clPercentage = CLPercentage; }
    static void setReevaluate( bool reevaluate )
      { _s_reevaluate = reevaluate; }

    virtual Generic_S_N<CNumeric>& operator=( const Generic_S_N<CNumeric>& rhs )
    {
      if ( this != &rhs )
      {
        // assert( typeid(rhs) == typeid(*this) );
         N::operator=( rhs ); 
        //Generic_S_NVD<CNumeric>::operator=( rhs ); 
      }  
      return *this;
    }
    
    Generic_S_NCLE<S, N, CNumeric>&
      operator=( const Generic_S_NCLE<S, N, CNumeric>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const N&>( rhs ) );
      return *this;
    }
};


// definition of static class elements -----------------------------------------

template <class S, class N, class CNumeric>
vector<N> Generic_S_NCLE<S, N, CNumeric>::_s_candidateList;

template <class S, class N, class CNumeric>
typename CNumeric::Range Generic_S_NCLE<S, N, CNumeric>::_s_candidateListPosition;

template <class S, class N, class CNumeric>
typename CNumeric::Range Generic_S_NCLE<S, N, CNumeric>::_s_remainingCandidateListTraversals;

template <class S, class N, class CNumeric>
double Generic_S_NCLE<S, N, CNumeric>::_s_clQuota=0.2;

template <class S, class N, class CNumeric>
double Generic_S_NCLE<S, N, CNumeric>::_s_clPercentage=0.5;

template <class S, class N, class CNumeric>
bool Generic_S_NCLE<S, N, CNumeric>::_s_reevaluate=true;


// definition of member functions ----------------------------------------------

template <class S, class N, class CNumeric> 
Generic_S_NCLE<S, N, CNumeric>
::Generic_S_NCLE( S *currentSolution,
                  NeighborhoodPosition itpos, short depth ) // =FirstNeighbor, =1
  : N( currentSolution, itpos, depth )
{ 
  if ( _s_remainingCandidateListTraversals > 0 )
  // use first entry of candidate list as first neighbor
  // (i.e. overwriting things done by the N constructor)
  {
    // cout << "CL-Iteration " << _remainingCandidateListTraversals << endl;
    assert( _s_candidateList.size() > 0 );
    _s_candidateListPosition = 0;
    *this = _s_candidateList[_s_candidateListPosition];

    if ( _s_reevaluate )
      evaluate( );
  }
  else
    if ( isValid( ) )
      _s_candidateList.push_back( *this );
}


template <class S, class N, class CNumeric> 
void Generic_S_NCLE<S, N, CNumeric>::operator++( )
{
  // cout << *this << **this << endl;

  if ( _s_remainingCandidateListTraversals > 0 )
  // get next entry of candidate list as next neighbor
  {
    ++_s_candidateListPosition;
    if ( _s_candidateListPosition >= _s_candidateList.size() )
    {
      N invalidNeighbor( _currentSolution, InvalidNeighbor );
      *this = invalidNeighbor;
      --_s_remainingCandidateListTraversals;
      if ( _s_remainingCandidateListTraversals == 0 )
        _s_candidateList.clear( );
    }
    else
    {
      S *currentSolution =  _currentSolution;
      *this = _s_candidateList[_s_candidateListPosition];
      _currentSolution = currentSolution; 

      if ( _s_reevaluate )
        evaluate( );
    }
  }
  else
  // construct next neighbor the usual way
  {
    N::operator++( );
    if ( isValid( ) )
      _s_candidateList.push_back( *this );
    else
      if ( !( _s_candidateList.empty() ) )
      // activate candidate list for next neighborhood sweep
      {
        random_shuffle( _s_candidateList.begin(), _s_candidateList.end() );
        sort( _s_candidateList.begin(), _s_candidateList.end(), betterEvaluation<N>() );
        vector<N>::iterator clIterator = _s_candidateList.begin();
        for( Range i=1; i<=_s_clQuota*_s_candidateList.size(); ++i )
          ++clIterator;
        _s_candidateList.erase( clIterator, _s_candidateList.end() );
        _s_remainingCandidateListTraversals = ceil( _s_clPercentage*_s_candidateList.size() );
      }  
  }
}

template <class S, class N, class CNumeric> 
ostream& operator<<( ostream& output, const Generic_S_NCLE<S, N, CNumeric>& neighbor )
{
  neighbor.print( output );
  return output;
}


// -- Generic_S_NCLA -----------------------------------------------------------
/*
   general mechanism for adapting all neighborhood classes derived from Generic_S_N
   - aspiration strategy
   idea: after one full neighborhood sweep the iterator checks only that many neighbors
         so that the quality of these neighbors is good enough with respect to some
         mean neighbor quality, which is adapted during the search;
         this strategy may be seen as a compromise between steepest descent and first descent
*/

// forward declaration
template <class S, class N, class CNumeric> class Generic_S_NCLA;


template <class S, class N, class CNumeric>
class Generic_S_NCLA
: public N
{
  public:
    typedef typename CNumeric::T T;
    typedef typename CNumeric::Range Range;

  protected:
    // if candidate list is active, neighbors are constructed such that after one neighbor with
    // quality _s_evaluationThreshold has been found _s_neighborhoodPercentage*neighborhoodSize
    // additional neighbors are constructed; _s_evaluationThreshold is computed as 
    // _s_averageEvaluation + _s_beta( _s_bestEvaluation - _s_averageEvaluation )
    static vector<T> _moveEvaluations;
    static T _s_averageEvaluation;
    static T _s_bestEvaluation;
    static T _s_evaluationThreshold;
    static Generic_S_NCLA<S, N, CNumeric> _s_nextNeighbor;
    static bool _s_NeighborhoodOverflow;
    static Range _s_status;
      // -4: initial state;
      // -3: CL strategy activated (parameters initialized, usual first neighbor to be used)
      // -2: CL strategy activated (_s_nextNeighbor to be used as first neighbor)
      // -1: CL neighborhood sweep active, but not yet a high quality neighbor found
      // >0: CL neighborhood sweep active, that many neighbors to be constructed
      //  0: CL neighborhood sweep finished 

    // parameters:
    static double _s_beta; // (0<=..<=1)
    static double _s_clPercentage; // (0<=..<=1)
    static double _s_smoothingParameter;  // exponential smoothing parameter for

  public:
    // constructor
    Generic_S_NCLA( S *currentSolution,
 		                NeighborhoodPosition itpos = FirstNeighbor,
                    short depth = 1 );

    Generic_S_NCLA( ) { ; }  // needed to define static member data _s_nextNeighbor

    Generic_S_NCLA( const Generic_S_NCLA<S, N, CNumeric>& original )
      : N( original )
    { ; }

    virtual Generic_S_N<CNumeric>* clone() const
      { return new Generic_S_NCLA<S, N, CNumeric>( *this ); }

    virtual void operator++( );

    void reset( ) 
    {
      _s_moveEvaluations.clear( );
      _s_status = -4;
    }

    static bool candidateListActive( )
      { return true; }

    static void setBeta( double beta )
      { _s_beta = beta; }
    static void setSmoothingParameter( double smoothingParameter )
      { _s_smoothingParameter = smoothingParameter; }
    static void setNeighborhoodPercentage( double CLPercentage )
      { _s_clPercentage = CLPercentage; }

    virtual Generic_S_N<CNumeric>& operator=( const Generic_S_N<CNumeric>& rhs )
    {
      if ( this != &rhs )
      {
        // assert( typeid(rhs) == typeid(*this) );
        N::operator=( rhs ); 
      }  
      return *this;
    }
    
    Generic_S_NCLA<S, N, CNumeric>&
      operator=( const Generic_S_NCLA<S, N, CNumeric>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const N&>( rhs ) ); 
      return *this;
    }
};


// definition of static class elements -----------------------------------------

template <class S, class N, class CNumeric>
vector<typename CNumeric::T> Generic_S_NCLA<S, N, CNumeric>::_moveEvaluations;

template <class S, class N, class CNumeric>
typename CNumeric::T Generic_S_NCLA<S, N, CNumeric>::_s_averageEvaluation;

template <class S, class N, class CNumeric>
typename CNumeric::T Generic_S_NCLA<S, N, CNumeric>::_s_bestEvaluation = numeric_limits<CNumeric::T>::min();

template <class S, class N, class CNumeric>
typename CNumeric::T Generic_S_NCLA<S, N, CNumeric>::_s_evaluationThreshold;

template <class S, class N, class CNumeric>
typename CNumeric::Range Generic_S_NCLA<S, N, CNumeric>::_s_status=-4;

template <class S, class N, class CNumeric>
Generic_S_NCLA<S, N, CNumeric> Generic_S_NCLA<S, N, CNumeric>::_s_nextNeighbor;

template <class S, class N, class CNumeric>
bool Generic_S_NCLA<S, N, CNumeric>::_s_NeighborhoodOverflow;

template <class S, class N, class CNumeric>
double Generic_S_NCLA<S, N, CNumeric>::_s_beta=0.5;

template <class S, class N, class CNumeric>
double Generic_S_NCLA<S, N, CNumeric>::_s_smoothingParameter=1;

template <class S, class N, class CNumeric>
double Generic_S_NCLA<S, N, CNumeric>::_s_clPercentage=0.1;


// definition of member functions ----------------------------------------------

template <class S, class N, class CNumeric> 
Generic_S_NCLA<S, N, CNumeric>
::Generic_S_NCLA( S *currentSolution,
                  NeighborhoodPosition itpos, short depth ) // =FirstNeighbor, =1
  : N( currentSolution, itpos, depth )
{ 
  if ( _s_status >= -3 )
  {
    getObserver()->comment( "new candidate list sweep" );

    if ( _s_status == -2 )
    // this is not the first restricted neighborhood sweep
    {
      S *currentSolution =  _currentSolution;
      *this = _s_nextNeighbor; // continue sweep at remembered position
      _currentSolution = currentSolution; 
      if ( isValid() )
        evaluate( );
    }
    // else keep neighbor

    if ( isValid( ) )
    {
      _moveEvaluations.push_back( **this );
      if ( **this > _s_bestEvaluation )
        _s_bestEvaluation = **this;

      if ( **this >= _s_evaluationThreshold )
        _s_status = _s_clPercentage*size();
      else
        _s_status = -1; // i.e. _s_evaluationThreshold not yet reached
    }
  }
  else
    // candidate list not yet active, so just gather information
    if ( isValid( ) )
    {
      _moveEvaluations.push_back( **this );
      if ( **this > _s_bestEvaluation )
        _s_bestEvaluation = **this;

      getObserver()->comment( "candidate list not yet active, gathering data ..." );
    }
    // else: empty neighborhood
}


template <class S, class N, class CNumeric> 
void Generic_S_NCLA<S, N, CNumeric>::operator++( )
{
  N::operator++( );

  // the implementation of this method is a little bit complicated,
  // so it may take some time to understand the following code

  if ( _s_status >= -3 )
  {
    if ( _s_status == 0 )
    // remember current neighbor and stop sweep
    {
      if ( isValid() )
      {
        _s_nextNeighbor = *this;
        *this = Generic_S_NCLA<S, N, CNumeric>( currentSolution(), InvalidNeighbor );
      }
      else
      {
        _s_nextNeighbor = N( currentSolution(), FirstNeighbor );

        // complete neighborhood sweep finished, so adapt parameters
        double newAverageEvaluation
               = ( accumulate( _moveEvaluations.begin(), _moveEvaluations.end(), 0 )
                 / T(_moveEvaluations.size()) );
        _s_averageEvaluation = _s_smoothingParameter*newAverageEvaluation + (1-_s_smoothingParameter)*_s_averageEvaluation;
        _s_evaluationThreshold = _s_averageEvaluation
                                + _s_beta*( _s_bestEvaluation - _s_averageEvaluation );
        _s_bestEvaluation = numeric_limits<T>::min();
        _moveEvaluations.clear();
      }
      _s_status = -2;
    }
    else
    // continue sweep, and gather information or adapt parameters
    {
      if ( _s_status > 0 )
        --_s_status;

      if ( !isValid( ) )
      {
        // complete neighborhood sweep finished, so adapt parameters
        double newAverageEvaluation
               = ( accumulate( _moveEvaluations.begin(), _moveEvaluations.end(), 0 )
                 / T(_moveEvaluations.size()) );
        _s_averageEvaluation = _s_smoothingParameter*newAverageEvaluation + (1-_s_smoothingParameter)*_s_averageEvaluation;
        _s_evaluationThreshold = _s_averageEvaluation
                                 + _s_beta*( _s_bestEvaluation - _s_averageEvaluation );
        _s_bestEvaluation = numeric_limits<T>::min();
        _moveEvaluations.clear();

        if ( _s_NeighborhoodOverflow ) 
        // reset candidate list
        {
          _s_status = -4;
          _s_NeighborhoodOverflow = false;
          _moveEvaluations.clear();
          getObserver()->comment( "neighborhood overflow, candidate list deactivated" );
        }
        else
        {
          _s_NeighborhoodOverflow = true;
          *this = N( currentSolution(), FirstNeighbor );
        }
      }

      if ( isValid( ) )
      {
        _moveEvaluations.push_back( **this );
        if ( **this > _s_bestEvaluation )
          _s_bestEvaluation = **this;

        if ( ( **this >= _s_evaluationThreshold ) && ( _s_status == -1 ) )
          // first time _s_evaluationThreshold reached in this sweep
          _s_status = _s_clPercentage*size();
      }
    }
  }
  else
  // candidate list not yet active, so construct next neighbor the usual way and gather information
  {
    if ( isValid( ) )
    {
      _moveEvaluations.push_back( **this );
      if ( **this > _s_bestEvaluation )
        _s_bestEvaluation = **this;
    }
    else
    {
      // activate candidate list and initialize parameters
      _s_status = -3;
      _s_averageEvaluation = ( accumulate( _moveEvaluations.begin(), _moveEvaluations.end(), 0 )
                               / T(_moveEvaluations.size()) );
      _s_evaluationThreshold = _s_averageEvaluation
                               + _s_beta*( _s_bestEvaluation - _s_averageEvaluation );
      _s_bestEvaluation = numeric_limits<T>::min();
      _moveEvaluations.clear();
      _s_NeighborhoodOverflow = false;
 
      getObserver()->comment( "candidate list activated from now on" );
    }  
  }
}


template <class S, class N, class CNumeric> 
ostream& operator<<( ostream& output, const Generic_S_NCLA<S, N, CNumeric>& neighbor )
{
  neighbor.print( output );
  return output;
}



}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif
