// Generic_S_NPilot: pilot transformation of neighborhoods

// ((this is experimental code))


#if !defined(__HotFrame_Generic_S_NPilot__)
#define __HotFrame_Generic_S_NPilot__

#include "basic.h"
#include "generic_s.h"
#include "search.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------
namespace HotFrame
{

// -- Generic_S_NPilot ---------------------------------------------------------
/*
   general mechanism for adapting all neighborhood classes derived from Generic_S_N
   idea: adapt move evalution by applying some pilotEvaluator, which is just a
         heuristic derived from Heuristic<C>, and using the evalution of the
         obtained solution
*/

// forward declaration
template <class S, class N, class CNumeric> class PilotNeighborhood;


template <class S, class N, class CNumeric>
class PilotNeighborhood
: public N
{
  public:
    typedef typename CNumeric::T T;
    typedef typename CNumeric::Range Range;

  public:
    static Heuristic<S> *_s_pilotEvaluator;

  public:
    // constructor
    PilotNeighborhood( S *currentSolution,
 		                   NeighborhoodPosition itpos = FirstNeighbor,
                       short depth = 1 )
    : N( currentSolution, itpos, depth )
    {
      if ( isValid() )
        evaluate();
    }

    PilotNeighborhood( const PilotNeighborhood<S, N, CNumeric>& original )
      : N( original )
    { ; }

    virtual void evaluate( )
    {
      if ( _s_pilotEvaluator )
      {
        if ( currentSolution()->observer()->active() )
        {
          currentSolution()->observer()->comment( "Pilot evaluation of neighbor:" );
          currentSolution()->observer()->currentNeighbor();
          currentSolution()->observer()->getStream() << *this;
          currentSolution()->observer()->flush();
        }
        
        S* pilotSolution = static_cast<S*>( currentSolution()->clone() );
        _deltaInformation = false;
        _evaluation = 0;
        pilotSolution->doMove( *this );
        _s_pilotEvaluator->search( *pilotSolution );
        _evaluation = currentSolution()->f() - pilotSolution->f(); // or: only "= - pilotSolution->f()" ?
        delete pilotSolution;
      }
      else
        _deltaInformation =
          currentSolution()->computeEvaluation( *this, _evaluation, _delta );
    }

    virtual Generic_S_N<CNumeric>& operator=( const Generic_S_N<CNumeric>& rhs )
    {
      if ( this != &rhs )
      {
        assert( typeid(rhs) == typeid(*this) );
         N::operator=( rhs ); 
      }  
      return *this;
    }
    
    PilotNeighborhood<S, N, CNumeric>&
      operator=( const PilotNeighborhood<S, N, CNumeric>& rhs )
    {
      if ( this != &rhs )
        operator=( static_cast<const N&>( rhs ) );
      return *this;
    }
};


template <class S, class N, class CNumeric>
Heuristic<S>* PilotNeighborhood<S, N, CNumeric>::_s_pilotEvaluator = 0;


template <class S, class N, class CNumeric> 
ostream& operator<<( ostream& output, const PilotNeighborhood<S, N, CNumeric>& neighbor )
{
  neighbor.print( output );
  return output;
}


}
//--- end of namespace HotFrame -----------------------------------------------
//------------------------------------------------------------------------------


#endif





