// sa_configs.h:
// defines specific heuristic configuration for simulated annealing and variations


#if !defined(__HotFrame_SA_CONFIG__)
#define __HotFrame_SA_CONFIG__

#include "sa.h"


//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

namespace HotFrame
{

//----------------------------------------------------------------------------
//-------------- Configuration of Simulated Annealing and Variations ---------
//----------------------------------------------------------------------------

template <class C>
class CClassicSimulatedAnnealing
{ 
  public:

    typedef typename C::CNumeric CNumeric;
    typedef typename C::Range Range;
    typedef typename C::T T;

    typedef typename C::S S;
    typedef typename C::N N;
	
    typedef C Cp;
    typedef CClassicSimulatedAnnealing<C> Configuration;
	
    typedef ClassicExponentialAcceptanceCriterion<Configuration> AcceptanceCriterion;
    
    typedef GeometricCooling<Configuration> CoolingSchedule;
    static double alpha;
    static double alpha1;
    static double alpha2;

    typedef ReheatingToHalfOfInitial<Configuration> Reheating;
};

template <class C>
double CClassicSimulatedAnnealing<C>::alpha = 0.95;

template <class C>
double CClassicSimulatedAnnealing<C>::alpha1 = 0.5;

template <class C>
double CClassicSimulatedAnnealing<C>::alpha2 = 0.5;



template <class C>
class CThresholdAccepting
{ 
  public:

    typedef typename C::CNumeric CNumeric;
    typedef typename C::Range Range;
    typedef typename C::T T;

    typedef typename C::S S;
    typedef typename C::N N;
	
    typedef C Cp;
    typedef CThresholdAccepting<C> Configuration;
	
    typedef ClassicThresholdAcceptanceCriterion<Configuration> AcceptanceCriterion;
    
    
    typedef GeometricCooling<Configuration> CoolingSchedule;
    static double alpha;
    static double alpha1;
    static double alpha2;

    typedef ReheatingToAverageOfBestAndInitial<Configuration> Reheating;
};



template <class C>
double CThresholdAccepting<C>::alpha = 0.95;

template <class C>
double CThresholdAccepting<C>::alpha1 = 0.5;

template <class C>
double CThresholdAccepting<C>::alpha2 = 0.5;




template <class C>
class CGreatDeluge
{ 
  public:

    typedef typename C::CNumeric CNumeric;
    typedef typename C::Range Range;
    typedef typename C::T T;

    typedef typename C::S S;
    typedef typename C::N N;
	
    typedef C Cp;
    typedef CGreatDeluge<C> Configuration;
	
    typedef AbsoluteThresholdAcceptanceCriterion<Configuration> AcceptanceCriterion;
    
    
    typedef GreatDelugeCooling<Configuration> CoolingSchedule;
    static double alpha;
    static double alpha1;
    static double alpha2;

    typedef ReheatingToBest<Configuration> Reheating;
};

template <class C>
double CGreatDeluge<C>::alpha = 0.95;

template <class C>
double CGreatDeluge<C>::alpha1 = 0.5;

template <class C>
double CGreatDeluge<C>::alpha2 = 0.5;




template <class C>
class CRecordToRecordTravel
{ 
  public:

    typedef typename C::CNumeric CNumeric;
    typedef typename C::Range Range;
    typedef typename C::T T;

    typedef typename C::S S;
    typedef typename C::N N;
	
    typedef C Cp;
    typedef CRecordToRecordTravel<C> Configuration;
	
    typedef AbsoluteThresholdAcceptanceCriterion<Configuration> AcceptanceCriterion;
    
    
    typedef RecordToRecordTravelCooling<Configuration> CoolingSchedule;
    static double alpha;
    static double alpha1;
    static double alpha2;

    typedef NoReheating<Configuration> Reheating;
};



template <class C>
double CRecordToRecordTravel<C>::alpha = 0.95;

template <class C>
double CRecordToRecordTravel<C>::alpha1 = 0.5;

template <class C>
double CRecordToRecordTravel<C>::alpha2 = 0.5;




//----------------------------------------------------------------------------
//-------------- Configuration of Simulated Annealing by Johnson et al. ------
//----------------------------------------------------------------------------

template <class C>
class CSimulatedAnnealingJetal
{ 
  public:

    typedef typename C::CNumeric CNumeric;
    typedef typename C::Range Range;
    typedef typename C::T T;

    typedef typename C::S S;
    typedef typename C::N N;
	
    typedef C Cp;
    typedef CSimulatedAnnealingJetal<C> Configuration;
	
    typedef ClassicExponentialAcceptanceCriterion<Configuration> AcceptanceCriterion;
    
    
    typedef GeometricCooling<Configuration> CoolingSchedule;
    static double alpha;
    static double alpha1;
    static double alpha2;

//    typedef ReheatingToHalfOfInitial<Configuration> Reheating;
    typedef ReheatingToAverageOfBestAndInitial<Configuration> Reheating;
};

template <class C>
double CSimulatedAnnealingJetal<C>::alpha = 0.95;

template <class C>
double CSimulatedAnnealingJetal<C>::alpha1 = 0.5;

template <class C>
double CSimulatedAnnealingJetal<C>::alpha2 = 0.5;



}
//----end of namespace HotFrame---------------------------------------
//--------------------------------------------------------------------

#endif
